/* 
 * Experiment 8: I2C Protocol - Control the LEDs
 * Find all the details about this experiment in the book << The Logic of things >>
 * Created 10 March 2017
 * By Zeroplus
 */

// The library Wire is necessary for using the I2C protocol
#include "Wire.h"

#define GREEN 0b00000110
#define RED 0b00000101

bool addrButton = 0, dataButton = 0;
volatile bool flagAddr = 0, flagData = 0, flagReadWriteButton = 0, flagSendButton = 0;
int address = 0x20, readWrite = 0, data = GREEN, addressOne = 0x20, addressTwo = 0x21;
const int bounceTime = 300; //300ms

void setup() {
  // Configure buttons as input pullup. Pressing a buttong ties it to Low, otherwise the pullup ties it to High.
  pinMode(9, INPUT_PULLUP);
  pinMode(8, INPUT_PULLUP);
  pinMode(2, INPUT_PULLUP);
  pinMode(A0, INPUT_PULLUP);
  setInterrupts();
  Serial.begin(9600);
  Wire.begin();
}

void loop() {
  /* The loop consists of four If statements that are true when a button has been pushed pressed.
     When a button is pressed, its variable is updated.
     When the button Send is pressed, a command is sent to the slave device.*/

  //Read Address button's status and update value 
  if (flagAddr == 1) {
    addrButton = !addrButton;
    if (addrButton == 0) {
      address = addressOne;
    }
    else {
      address = addressTwo;
    }
    eliminate_bounce();
    flagAddr = 0;
  }
  
  //Read Data button's status and update value
  if (flagData == 1) {
     dataButton = !dataButton;
    if (dataButton == 0) {
      data = GREEN;
    }
    else {
      data = RED;
    }
    eliminate_bounce();
    flagData = 0;
  }

  //Read R/W button's status and update value
  if (flagReadWriteButton == 1) {
    readWrite = !readWrite;
    eliminate_bounce();
    flagReadWriteButton = 0;
  }

  //Read Send button's status and send a command to the LEDs module if the button has been pressed
  if (flagSendButton == 1) {
    Wire.beginTransmission(address);
    if (readWrite == 1) {
      Wire.requestFrom(address, 4);
      while (Wire.available()) {
        int c = Wire.read();
        Serial.print(c);
      }
      Serial.print("\n");
    }
    else {
      Wire.write(data);
    }
    Wire.endTransmission();
    eliminate_bounce();
    flagSendButton = 0;
  }
}

// The function setInterrupts() sets the correct values in the register of the microcontroller in order to configure the interrupts
// Details can be found in the data sheet of the ATMega 328
void setInterrupts() {
  cli();                // pause all interrupts
  PCICR = 0x07;         // enable PCIE0, PCIE1, PCIE2
  PCMSK0 = 0b00000011;  // enable PCINT0 and PCINT1 -> pin 8 and pin 9
  PCMSK1 = 1 << 0;      // enable PCINT8  -> pin A0
  PCMSK2 = 0b00000100;  // enable PCINT18  -> pin 2
  sei();                // turn on all interrupts
}

// Below the three interrupt handlers update the flag status of each button
// Note: Pin 8 and pin 9 share the same interrupt handler
ISR(PCINT0_vect) {
  bool pin8State = 0, pin9State = 0;
  pin8State = digitalRead(8);
  pin9State = digitalRead(9);
  //Checking which pin 8 or 9 triggered the interrupt
  if (pin8State == 0) {
    flagAddr = 1;
  }
  else if (pin9State == 0) {
    flagSendButton = 1;
  }
  else {
    //do nothing
  }
}
ISR(PCINT1_vect) {
  flagData = 1;
}
ISR(PCINT2_vect) {
  flagReadWriteButton = 1;
}

// Whenever a button is pressed the contact is not prefect and results in fast ON/OFF transitions for a short period of time.
// The function eleminate_bounce() runs long enough for the transitions to stop.
void eliminate_bounce() {
  unsigned long current_time = 0;
  unsigned long start_time = millis();
  do {
    current_time = millis();
  } while (current_time - start_time <= bounceTime);
}
